"""Create discord messages from killmails."""

import dhooks_lite
from requests.exceptions import HTTPError

from eveuniverse.helpers import EveEntityNameResolver
from eveuniverse.models import EveEntity, EveSolarSystem

from allianceauth.eveonline.evelinks import dotlan, eveimageserver, zkillboard
from allianceauth.services.hooks import get_extension_logger
from app_utils.django import app_labels
from app_utils.logging import LoggerAddTag
from app_utils.urls import static_file_absolute_url
from app_utils.views import humanize_value

from .. import __title__
from ..models import Tracker
from .killmails import ZKB_KILLMAIL_BASEURL, Killmail

ICON_SIZE = 128


logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def create_content(tracker: Tracker, intro_text: str = None) -> str:
    """Create content for Discord message for a killmail."""

    intro_parts = []

    if tracker.ping_type == Tracker.ChannelPingType.EVERYBODY:
        intro_parts.append("@everybody")
    elif tracker.ping_type == Tracker.ChannelPingType.HERE:
        intro_parts.append("@here")

    if tracker.ping_groups.exists():
        if "discord" in app_labels():
            DiscordUser = _import_discord_user()
            for group in tracker.ping_groups.all():
                try:
                    role = DiscordUser.objects.group_to_role(group)
                except HTTPError:
                    logger.warning(
                        "Failed to get Discord roles. Can not ping groups.",
                        exc_info=True,
                    )
                else:
                    if role:
                        intro_parts.append(f"<@&{role['id']}>")

        else:
            logger.warning(
                "Discord service needs to be installed in order "
                "to use groups ping features."
            )

    if tracker.is_posting_name:
        intro_parts.append(f"Tracker **{tracker.name}**:")

    intro_parts_2 = []
    if intro_text:
        intro_parts_2.append(intro_text)
    if intro_parts:
        intro_parts_2.append(" ".join(intro_parts))

    return "\n".join(intro_parts_2)


def _import_discord_user():
    from allianceauth.services.modules.discord.models import DiscordUser

    return DiscordUser


def create_embed(tracker: Tracker, killmail: Killmail) -> dhooks_lite.Embed:
    """Create Discord embed for a killmail."""

    resolver = EveEntity.objects.bulk_resolve_names(ids=killmail.entity_ids())

    # victim
    if killmail.victim.alliance_id:
        victim_organization = resolver.to_name(killmail.victim.alliance_id)
        victim_org_url = zkillboard.alliance_url(killmail.victim.alliance_id)
        victim_org_icon_url = eveimageserver.alliance_logo_url(
            killmail.victim.alliance_id, size=ICON_SIZE
        )
    elif killmail.victim.corporation_id:
        victim_organization = resolver.to_name(killmail.victim.corporation_id)
        victim_org_url = zkillboard.corporation_url(killmail.victim.corporation_id)
        victim_org_icon_url = eveimageserver.corporation_logo_url(
            killmail.victim.corporation_id, size=ICON_SIZE
        )
    else:
        victim_organization = None
        victim_org_url = None
        victim_org_icon_url = None

    if killmail.victim.corporation_id:
        victim_corporation_zkb_link = _corporation_zkb_link(
            tracker, killmail.victim.corporation_id, resolver
        )
    else:
        victim_corporation_zkb_link = ""

    if killmail.victim.character_id:
        victim_character_zkb_link = _character_zkb_link(
            tracker, killmail.victim.character_id, resolver
        )
        victim_str = f"{victim_character_zkb_link} ({victim_corporation_zkb_link})"
    elif killmail.victim.corporation_id:
        victim_str = victim_corporation_zkb_link
    else:
        victim_str = ""

    # final attacker
    for attacker in killmail.attackers:
        if attacker.is_final_blow:
            final_attacker = attacker
            break
    else:
        final_attacker = None

    if final_attacker:
        if final_attacker.corporation_id:
            final_attacker_corporation_zkb_link = _corporation_zkb_link(
                tracker, final_attacker.corporation_id, resolver
            )
        else:
            final_attacker_corporation_zkb_link = ""

        if final_attacker.character_id and final_attacker.corporation_id:
            final_attacker_character_zkb_link = _character_zkb_link(
                tracker, final_attacker.character_id, resolver
            )
            final_attacker_str = (
                f"{final_attacker_character_zkb_link} "
                f"({final_attacker_corporation_zkb_link})"
            )
        elif final_attacker.corporation_id:
            final_attacker_str = f"{final_attacker_corporation_zkb_link}"
        elif final_attacker.faction_id:
            final_attacker_str = f"**{resolver.to_name(final_attacker.faction_id)}**"
        else:
            final_attacker_str = "(Unknown final_attacker)"

        final_attacker_ship_type_name = resolver.to_name(final_attacker.ship_type_id)

    else:
        final_attacker_str = ""
        final_attacker_ship_type_name = ""

    if killmail.solar_system_id:
        solar_system, _ = EveSolarSystem.objects.get_or_create_esi(
            id=killmail.solar_system_id
        )
        solar_system_link = tracker.webhook.create_message_link(
            name=solar_system.name, url=dotlan.solar_system_url(solar_system.name)
        )
        region_name = solar_system.eve_constellation.eve_region.name
        solar_system_text = f"{solar_system_link} ({region_name})"
    else:
        solar_system_text = ""

    # self info
    show_as_fleetkill = False
    distance_text = ""
    main_org_text = ""
    main_org_name = ""
    main_org_icon_url = eveimageserver.alliance_logo_url(1, size=ICON_SIZE)
    main_ship_group_text = ""
    tracked_ship_types_text = ""
    embed_color = None
    if killmail.tracker_info:
        show_as_fleetkill = tracker.identify_fleets
        embed_color = int(tracker.color[1:], 16) if tracker.color else None
        if tracker.origin_solar_system:
            origin_solar_system_link = tracker.webhook.create_message_link(
                name=tracker.origin_solar_system.name,
                url=dotlan.solar_system_url(tracker.origin_solar_system.name),
            )
            if killmail.tracker_info.distance is not None:
                distance_str = f"{killmail.tracker_info.distance:,.1f}"
            else:
                distance_str = "?"

            if killmail.tracker_info.jumps is not None:
                jumps_str = killmail.tracker_info.jumps
            else:
                jumps_str = "?"

            distance_text = (
                f"\nDistance from {origin_solar_system_link}: "
                f"{distance_str} LY | {jumps_str} jumps"
            )

        # main group
        main_org = killmail.tracker_info.main_org
        if main_org:
            main_org_name = resolver.to_name(main_org.id)
            if main_org.is_corporation:
                main_org_link = _corporation_zkb_link(tracker, main_org.id, resolver)
                main_org_icon_url = eveimageserver.corporation_logo_url(
                    main_org.id, size=ICON_SIZE
                )
            else:
                main_org_link = _alliance_zkb_link(tracker, main_org.id, resolver)
                main_org_icon_url = eveimageserver.alliance_logo_url(
                    main_org.id, size=ICON_SIZE
                )
            main_org_text = f" | Main group: {main_org_link} ({main_org.count})"
        else:
            show_as_fleetkill = False

        # main ship group
        main_ship_group = killmail.tracker_info.main_ship_group
        if main_ship_group:
            main_ship_group_text = f"\nMain ship class: **{main_ship_group.name}**"

        # tracked attacker ships
        matching_ship_type_ids = killmail.tracker_info.matching_ship_type_ids
        if matching_ship_type_ids:
            ship_types_text = "**, **".join(
                sorted(
                    [resolver.to_name(type_id) for type_id in matching_ship_type_ids]
                )
            )
            tracked_ship_types_text = (
                f"\nTracked ship types involved: **{ship_types_text}**"
            )

    victim_ship_type_name = resolver.to_name(killmail.victim.ship_type_id)
    total_value = (
        humanize_value(killmail.zkb.total_value) if killmail.zkb.total_value else "?"
    )
    description = (
        f"{victim_str} lost their **{victim_ship_type_name}** "
        f"in {solar_system_text} "
        f"worth **{total_value}** ISK.\n"
        f"Final blow by {final_attacker_str} "
        f"in a **{final_attacker_ship_type_name}**.\n"
        f"Attackers: **{len(killmail.attackers):,}**{main_org_text}"
        f"{main_ship_group_text}"
        f"{tracked_ship_types_text}"
        f"{distance_text}"
    )
    solar_system_name = resolver.to_name(killmail.solar_system_id)
    if show_as_fleetkill:
        title = f"{solar_system_name} | {main_org_name} | Fleetkill"
    else:
        title = f"{solar_system_name} | {victim_ship_type_name} | Killmail"

    if show_as_fleetkill:
        thumbnail_url = main_org_icon_url
    else:
        thumbnail_url = eveimageserver.type_icon_url(
            killmail.victim.ship_type_id, size=ICON_SIZE
        )

    zkb_killmail_url = f"{ZKB_KILLMAIL_BASEURL}{killmail.id}/"
    # TODO This is a workaround for Embed.Author.name. Address in dhooks_lite
    author = (
        dhooks_lite.Author(
            name=victim_organization if victim_organization else "?",
            url=victim_org_url,
            icon_url=victim_org_icon_url,
        )
        if victim_organization and victim_org_url and victim_org_icon_url
        else None
    )
    zkb_icon_url = static_file_absolute_url("killtracker/zkb_icon.png")
    embed = dhooks_lite.Embed(
        author=author,
        description=description,
        title=title,
        url=zkb_killmail_url,
        thumbnail=dhooks_lite.Thumbnail(url=thumbnail_url),
        footer=dhooks_lite.Footer(text="zKillboard", icon_url=zkb_icon_url),
        timestamp=killmail.time,
        color=embed_color,
    )
    return embed


def _character_zkb_link(
    tracker, entity_id: int, resolver: EveEntityNameResolver
) -> str:
    return tracker.webhook.create_message_link(
        name=resolver.to_name(entity_id), url=zkillboard.character_url(entity_id)
    )


def _corporation_zkb_link(
    tracker, entity_id: int, resolver: EveEntityNameResolver
) -> str:
    return tracker.webhook.create_message_link(
        name=resolver.to_name(entity_id), url=zkillboard.corporation_url(entity_id)
    )


def _alliance_zkb_link(tracker, entity_id: int, resolver: EveEntityNameResolver) -> str:
    return tracker.webhook.create_message_link(
        name=resolver.to_name(entity_id), url=zkillboard.alliance_url(entity_id)
    )



from unittest.mock import patch

import dhooks_lite
import requests_mock

from django.core.cache import cache
from django.test import TestCase
from django.test.utils import override_settings

from killtracker import tasks
from killtracker.core.killmails import ZKB_REDISQ_URL

from .testdata.factories import TrackerFactory
from .testdata.helpers import LoadTestDataMixin, killmails_data

PACKAGE_PATH = "killtracker"


@patch("celery.app.task.Context.called_directly", False)  # make retry work with eager
@override_settings(CELERY_ALWAYS_EAGER=True)
@patch(PACKAGE_PATH + ".tasks.is_esi_online", lambda: True)
@patch(PACKAGE_PATH + ".models.dhooks_lite.Webhook.execute", spec=True)
@requests_mock.Mocker()
class TestTasksEnd2End(LoadTestDataMixin, TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cache.clear()
        cls.tracker_1 = TrackerFactory(
            name="My Tracker",
            exclude_null_sec=True,
            exclude_w_space=True,
            webhook=cls.webhook_1,
        )

    @patch(PACKAGE_PATH + ".tasks.retry_task_if_esi_is_down", lambda x: None)
    def test_normal_case(self, requests_mocker, mock_execute):
        # given
        mock_execute.return_value = dhooks_lite.WebhookResponse(dict(), status_code=200)
        requests_mocker.register_uri(
            "GET",
            ZKB_REDISQ_URL,
            [
                {"status_code": 200, "json": {"package": killmails_data()[10000001]}},
                {"status_code": 200, "json": {"package": killmails_data()[10000002]}},
                {"status_code": 200, "json": {"package": killmails_data()[10000003]}},
                {"status_code": 200, "json": {"package": None}},
            ],
        )
        # when
        tasks.run_killtracker.delay()
        # then
        self.assertEqual(mock_execute.call_count, 2)
        _, kwargs = mock_execute.call_args_list[0]
        self.assertIn("My Tracker", kwargs["content"])
        self.assertIn("10000001", kwargs["embeds"][0].url)
        _, kwargs = mock_execute.call_args_list[1]
        self.assertIn("My Tracker", kwargs["content"])
        self.assertIn("10000002", kwargs["embeds"][0].url)

    @patch(PACKAGE_PATH + ".tasks.retry_task_if_esi_is_down")
    def test_should_retry_when_esi_error_limit_reached(
        self, requests_mocker, mock_retry_task_if_esi_is_down, mock_execute
    ):
        def my_retry_task_if_esi_is_down(task):
            """Retry the task one time only."""
            if task.request.retries < 1:
                raise task.retry()

        # given
        mock_execute.return_value = dhooks_lite.WebhookResponse(dict(), status_code=200)
        mock_retry_task_if_esi_is_down.side_effect = my_retry_task_if_esi_is_down
        requests_mocker.register_uri(
            "GET",
            ZKB_REDISQ_URL,
            [
                {"status_code": 200, "json": {"package": killmails_data()[10000001]}},
                {"status_code": 200, "json": {"package": None}},
            ],
        )
        # when
        tasks.run_killtracker.delay()
        # then
        self.assertEqual(mock_execute.call_count, 1)



from django.contrib.auth.models import User
from django.urls import reverse
from django_webtest import WebTest
from eveuniverse.models import EveType

from allianceauth.eveonline.models import EveCorporationInfo
from app_utils.testing import create_fake_user

from killtracker.models import Tracker, Webhook

from .testdata.factories import TrackerFactory
from .testdata.helpers import LoadTestDataMixin
from .testdata.load_eveuniverse import load_eveuniverse


class TestTrackerChangeList(LoadTestDataMixin, WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_superuser(
            "Bruce_Wayne", "bruce@example.com", "password"
        )

    def setUp(self) -> None:
        TrackerFactory(webhook=self.webhook_1, exclude_high_sec=True)
        TrackerFactory(
            webhook=self.webhook_1,
            origin_solar_system_id=30003067,
            require_max_jumps=3,
        )
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2001)
        )

    def test_can_open_page_normally(self):
        # login
        self.app.set_user(self.user)

        # user tries to add new notification rule
        add_page = self.app.get(reverse("admin:killtracker_tracker_changelist"))
        self.assertEqual(add_page.status_code, 200)


class TestTrackerValidations(LoadTestDataMixin, WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.webhook = Webhook.objects.create(name="Dummy", url="http://www.example.com")
        cls.user = User.objects.create_superuser(
            "Bruce_Wayne", "bruce@example.com", "password"
        )
        cls.url_add = reverse("admin:killtracker_tracker_add")
        cls.url_changelist = reverse("admin:killtracker_tracker_changelist")

    def setUp(self) -> None:
        Tracker.objects.all().delete()

    def _open_page(self) -> object:
        # login
        self.app.set_user(self.user)

        # user tries to add new notification rule
        add_page = self.app.get(self.url_add)
        self.assertEqual(add_page.status_code, 200)
        form = add_page.form
        form["name"] = "Test Tracker"
        form["webhook"] = self.webhook_1.pk
        return form

    def test_no_errors(self):
        form = self._open_page()
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, self.url_changelist)
        self.assertEqual(Tracker.objects.count(), 1)

    def test_can_not_have_require_max_jumps_without_origin(self):
        form = self._open_page()
        form["require_max_jumps"] = 10
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_can_not_have_require_max_distance_without_origin(self):
        form = self._open_page()
        form["require_max_distance"] = 5
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_can_not_have_same_options_attacker_alliances(self):
        form = self._open_page()
        form["exclude_attacker_alliances"] = [
            self.alliance_3001.pk,
            self.alliance_3011.pk,
        ]
        form["require_attacker_alliances"] = [self.alliance_3001.pk]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_can_not_have_same_options_attacker_corporations(self):
        form = self._open_page()
        form["exclude_attacker_corporations"] = [
            self.corporation_2001.pk,
            self.corporation_2011.pk,
        ]
        form["require_attacker_corporations"] = [self.corporation_2011.pk]
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_min_attackers_must_be_less_than_max_attackers(self):
        form = self._open_page()
        form["require_min_attackers"] = 10
        form["require_max_attackers"] = 5
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_can_not_exclude_all_space_types(self):
        form = self._open_page()
        form["exclude_high_sec"] = True
        form["exclude_low_sec"] = True
        form["exclude_null_sec"] = True
        form["exclude_w_space"] = True
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)

    def test_can_both_exclude_and_require_npc_kills(self):
        form = self._open_page()
        form["exclude_npc_kills"] = True
        form["require_npc_kills"] = True
        response = form.submit()

        # assert results
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please correct the error below", response.text)
        self.assertEqual(Tracker.objects.count(), 0)


class TestTrackerChangeForm(WebTest):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = create_fake_user(1001, "Bruce Wayne")
        cls.user.is_staff = True
        cls.user.is_superuser = True
        cls.user.save()
        load_eveuniverse()
        cls.npc_type = EveType.objects.get(id=23320)

    def test_do_not_show_npcs_when_not_used_and_not_enabled(self):
        # given
        tracker = TrackerFactory()
        url = reverse("admin:killtracker_tracker_change", args=[tracker.id])
        self.app.set_user(self.user)
        # when
        response = self.app.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertNotIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertNotIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)

    def test_show_npcs_when_in_use(self):
        # given
        tracker = TrackerFactory()
        tracker.require_attackers_ship_types.add(self.npc_type)
        url = reverse("admin:killtracker_tracker_change", args=[tracker.id])
        self.app.set_user(self.user)
        # when
        response = self.app.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)

    def test_show_npcs_when_enabled(self):
        # given
        tracker = TrackerFactory()
        tracker.require_attackers_ship_types.add(self.npc_type)
        url = reverse("admin:killtracker_tracker_change", args=[tracker.id])
        self.app.set_user(self.user)
        self.app.session["killtracker_toogle_npc"] = True
        # when
        response = self.app.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)

    def test_can_enable_npcs(self):
        # given
        tracker = TrackerFactory()
        url = reverse("admin:killtracker_tracker_change", args=[tracker.id])
        self.app.set_user(self.user)
        initial_page = self.app.get(url)
        # when
        response = initial_page.click(linkid="toogle-npc").follow()
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)


class TestTrackerAddForm(WebTest):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = create_fake_user(1001, "Bruce Wayne")
        cls.user.is_staff = True
        cls.user.is_superuser = True
        cls.user.save()
        load_eveuniverse()
        cls.npc_type = EveType.objects.get(id=23320)

    def test_do_not_show_npcs_when_not_enabled(self):
        # given
        url = reverse("admin:killtracker_tracker_add")
        self.app.set_user(self.user)
        # when
        response = self.app.get(url)
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertNotIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertNotIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)

    def test_can_enable_npcs(self):
        # given
        url = reverse("admin:killtracker_tracker_add")
        self.app.set_user(self.user)
        initial_page = self.app.get(url)
        # when
        response = initial_page.click(linkid="toogle-npc").follow()
        # then
        self.assertEqual(response.status_code, 200)
        form = response.form
        require_attackers_ship_types_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_types"].options
        }
        self.assertIn(self.npc_type.id, require_attackers_ship_types_ids)
        require_attackers_ship_groups_ids = {
            int(obj[0]) for obj in form["require_attackers_ship_groups"].options
        }
        self.assertIn(self.npc_type.eve_group_id, require_attackers_ship_groups_ids)



import json
from datetime import timedelta
from unittest.mock import patch

import dhooks_lite
import requests_mock
from bs4 import BeautifulSoup
from markdown import markdown

from django.core.cache import cache
from django.test import TestCase
from django.utils.timezone import now
from eveuniverse.models import (
    EveConstellation,
    EveEntity,
    EveRegion,
    EveSolarSystem,
    EveType,
)

from app_utils.json import JSONDateTimeDecoder
from app_utils.testing import NoSocketsTestCase

from killtracker.core.killmails import Killmail
from killtracker.exceptions import WebhookTooManyRequests
from killtracker.models import EveKillmail, Webhook

from .testdata.factories import TrackerFactory
from .testdata.helpers import LoadTestDataMixin, load_eve_killmails, load_killmail

MODELS_PATH = "killtracker.models"


class TestWebhookQueue(LoadTestDataMixin, TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    def setUp(self) -> None:
        self.webhook_1.main_queue.clear()
        self.webhook_1.error_queue.clear()

    def test_reset_failed_messages(self):
        message = "Test message"
        self.webhook_1.error_queue.enqueue(message)
        self.webhook_1.error_queue.enqueue(message)
        self.assertEqual(self.webhook_1.error_queue.size(), 2)
        self.assertEqual(self.webhook_1.main_queue.size(), 0)
        self.webhook_1.reset_failed_messages()
        self.assertEqual(self.webhook_1.error_queue.size(), 0)
        self.assertEqual(self.webhook_1.main_queue.size(), 2)

    def test_discord_message_asjson_normal(self):
        embed = dhooks_lite.Embed(description="my_description")
        result = Webhook._discord_message_asjson(
            content="my_content",
            username="my_username",
            avatar_url="my_avatar_url",
            embeds=[embed],
        )
        message_python = json.loads(result, cls=JSONDateTimeDecoder)
        expected = {
            "content": "my_content",
            "embeds": [{"description": "my_description", "type": "rich"}],
            "username": "my_username",
            "avatar_url": "my_avatar_url",
        }
        self.assertDictEqual(message_python, expected)

    def test_discord_message_asjson_empty(self):
        with self.assertRaises(ValueError):
            Webhook._discord_message_asjson("")


class TestEveKillmailManager(LoadTestDataMixin, NoSocketsTestCase):
    def test_create_from_killmail(self):
        # given
        killmail = load_killmail(10000001)
        # when
        eve_killmail = EveKillmail.objects.create_from_killmail(killmail)
        # then
        self.assertIsInstance(eve_killmail, EveKillmail)
        self.assertEqual(eve_killmail.id, 10000001)
        self.assertEqual(eve_killmail.solar_system, EveEntity.objects.get(id=30004984))
        self.assertAlmostEqual(eve_killmail.time, now(), delta=timedelta(seconds=60))

        self.assertEqual(eve_killmail.alliance, EveEntity.objects.get(id=3011))
        self.assertEqual(eve_killmail.character, EveEntity.objects.get(id=1011))
        self.assertEqual(eve_killmail.corporation, EveEntity.objects.get(id=2011))
        self.assertEqual(eve_killmail.faction, EveEntity.objects.get(id=500004))
        self.assertEqual(eve_killmail.damage_taken, 434)
        self.assertEqual(eve_killmail.ship_type, EveEntity.objects.get(id=603))

        attacker_ids = list(eve_killmail.attackers.values_list("pk", flat=True))
        self.assertEqual(len(attacker_ids), 3)

        attacker = eve_killmail.attackers.get(pk=attacker_ids[0])
        self.assertEqual(attacker.alliance, EveEntity.objects.get(id=3001))
        self.assertEqual(attacker.character, EveEntity.objects.get(id=1001))
        self.assertEqual(attacker.corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(attacker.faction, EveEntity.objects.get(id=500001))
        self.assertEqual(attacker.damage_done, 434)
        self.assertEqual(attacker.security_status, -10)
        self.assertEqual(attacker.ship_type, EveEntity.objects.get(id=34562))
        self.assertEqual(attacker.weapon_type, EveEntity.objects.get(id=2977))
        self.assertTrue(attacker.is_final_blow)

        attacker = eve_killmail.attackers.get(pk=attacker_ids[1])
        self.assertEqual(attacker.alliance, EveEntity.objects.get(id=3001))
        self.assertEqual(attacker.character, EveEntity.objects.get(id=1002))
        self.assertEqual(attacker.corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(attacker.faction, EveEntity.objects.get(id=500001))
        self.assertEqual(attacker.damage_done, 50)
        self.assertEqual(attacker.security_status, -10)
        self.assertEqual(attacker.ship_type, EveEntity.objects.get(id=3756))
        self.assertEqual(attacker.weapon_type, EveEntity.objects.get(id=2488))
        self.assertFalse(attacker.is_final_blow)

        attacker = eve_killmail.attackers.get(pk=attacker_ids[2])
        self.assertEqual(attacker.alliance, EveEntity.objects.get(id=3001))
        self.assertEqual(attacker.character, EveEntity.objects.get(id=1003))
        self.assertEqual(attacker.corporation, EveEntity.objects.get(id=2001))
        self.assertEqual(attacker.faction, EveEntity.objects.get(id=500001))
        self.assertEqual(attacker.damage_done, 99)
        self.assertEqual(attacker.security_status, 5)
        self.assertEqual(attacker.ship_type, EveEntity.objects.get(id=3756))
        self.assertEqual(attacker.weapon_type, EveEntity.objects.get(id=2488))
        self.assertFalse(attacker.is_final_blow)

        self.assertEqual(eve_killmail.location_id, 50012306)
        self.assertEqual(eve_killmail.fitted_value, 10000)
        self.assertEqual(eve_killmail.total_value, 10000)
        self.assertEqual(eve_killmail.zkb_points, 1)
        self.assertFalse(eve_killmail.is_npc)
        self.assertFalse(eve_killmail.is_solo)
        self.assertFalse(eve_killmail.is_awox)

    def test_update_or_create_from_killmail(self):
        killmail = load_killmail(10000001)

        # first time will be created
        eve_killmail, created = EveKillmail.objects.update_or_create_from_killmail(
            killmail
        )
        self.assertTrue(created)
        self.assertEqual(eve_killmail.solar_system_id, 30004984)

        # update record
        eve_killmail.solar_system = EveEntity.objects.get(id=30045349)
        eve_killmail.save()
        eve_killmail.refresh_from_db()
        self.assertEqual(eve_killmail.solar_system_id, 30045349)

        # 2nd time will be updated
        eve_killmail, created = EveKillmail.objects.update_or_create_from_killmail(
            killmail
        )
        self.assertEqual(eve_killmail.id, 10000001)
        self.assertFalse(created)
        self.assertEqual(eve_killmail.solar_system_id, 30004984)

    @patch("killtracker.managers.KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS", 1)
    def test_delete_stale(self):
        load_eve_killmails([10000001, 10000002, 10000003])
        km = EveKillmail.objects.get(id=10000001)
        km.time = now() - timedelta(days=1, seconds=1)
        km.save()

        _, details = EveKillmail.objects.delete_stale()

        self.assertEqual(details["killtracker.EveKillmail"], 1)
        self.assertEqual(EveKillmail.objects.count(), 2)
        self.assertTrue(EveKillmail.objects.filter(id=10000002).exists())
        self.assertTrue(EveKillmail.objects.filter(id=10000003).exists())

    @patch("killtracker.managers.KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS", 0)
    def test_dont_delete_stale_when_turned_off(self):
        load_eve_killmails([10000001, 10000002, 10000003])
        km = EveKillmail.objects.get(id=10000001)
        km.time = now() - timedelta(days=1, seconds=1)
        km.save()

        self.assertIsNone(EveKillmail.objects.delete_stale())
        self.assertEqual(EveKillmail.objects.count(), 3)

    def test_load_entities(self):
        load_eve_killmails([10000001, 10000002])
        self.assertEqual(EveKillmail.objects.all().load_entities(), 0)


class TestHasLocalizationClause(LoadTestDataMixin, NoSocketsTestCase):
    def test_has_localization_filter_1(self):
        tracker = TrackerFactory.build(webhook=self.webhook_1, exclude_high_sec=True)
        self.assertTrue(tracker.has_localization_clause)

        tracker = TrackerFactory.build(webhook=self.webhook_1, exclude_low_sec=True)
        self.assertTrue(tracker.has_localization_clause)

        tracker = TrackerFactory.build(webhook=self.webhook_1, exclude_null_sec=True)
        self.assertTrue(tracker.has_localization_clause)

        tracker = TrackerFactory.build(webhook=self.webhook_1, exclude_w_space=True)
        self.assertTrue(tracker.has_localization_clause)

        tracker = TrackerFactory.build(webhook=self.webhook_1, require_max_distance=10)
        self.assertTrue(tracker.has_localization_clause)

        tracker = TrackerFactory.build(webhook=self.webhook_1, require_max_jumps=10)
        self.assertTrue(tracker.has_localization_clause)

    def test_has_no_matching_clause(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        self.assertFalse(tracker.has_localization_clause)

    def test_has_localization_filter_3(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_regions.add(EveRegion.objects.get(id=10000014))
        self.assertTrue(tracker.has_localization_clause)

    def test_has_localization_filter_4(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_constellations.add(EveConstellation.objects.get(id=20000169))
        self.assertTrue(tracker.has_localization_clause)

    def test_has_localization_filter_5(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_solar_systems.add(EveSolarSystem.objects.get(id=30001161))
        self.assertTrue(tracker.has_localization_clause)


class TestHasTypeClause(LoadTestDataMixin, NoSocketsTestCase):
    def test_has_no_matching_clause(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        self.assertFalse(tracker.has_type_clause)

    def test_has_require_attackers_ship_groups(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attackers_ship_groups.add(self.type_svipul.eve_group)
        self.assertTrue(tracker.has_type_clause)

    def test_has_require_attackers_ship_types(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attackers_ship_types.add(self.type_svipul)
        self.assertTrue(tracker.has_type_clause)

    def test_has_require_victim_ship_groups(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_victim_ship_groups.add(self.type_svipul.eve_group)
        self.assertTrue(tracker.has_type_clause)

    def test_has_require_victim_ship_types(self):
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_victim_ship_types.add(self.type_svipul)
        self.assertTrue(tracker.has_type_clause)


class TestSaveMethod(LoadTestDataMixin, NoSocketsTestCase):
    def test_black_color_is_none(self):
        tracker = TrackerFactory(webhook=self.webhook_1, color="#000000")
        tracker.refresh_from_db()
        self.assertFalse(tracker.color)


@requests_mock.Mocker()
class TestWebhookSendMessage(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.message = Webhook._discord_message_asjson(content="Test message")
        cache.clear()

    def test_when_send_ok_returns_true(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "POST",
            self.webhook_1.url,
            status_code=200,
            json={
                "name": "test webhook",
                "type": 1,
                "channel_id": "199737254929760256",
                "token": "3d89bb7572e0fb30d8128367b3b1b44fecd1726de135cbe28a41f8b2f777c372ba2939e72279b94526ff5d1bd4358d65cf11",
                "avatar": None,
                "guild_id": "199737254929760256",
                "id": "223704706495545344",
                "application_id": None,
                "user": {
                    "username": "test",
                    "discriminator": "7479",
                    "id": "190320984123768832",
                    "avatar": "b004ec1740a63ca06ae2e14c5cee11f3",
                    "public_flags": 131328,
                },
            },
        )
        # when
        response = self.webhook_1.send_message_to_webhook(self.message)
        # then
        self.assertTrue(response.status_ok)
        self.assertTrue(requests_mocker.called)

    def test_when_send_not_ok_returns_false(self, requests_mocker):
        # given
        requests_mocker.register_uri("POST", self.webhook_1.url, status_code=404)
        # when
        response = self.webhook_1.send_message_to_webhook(self.message)
        # then
        self.assertFalse(response.status_ok)
        self.assertTrue(requests_mocker.called)

    def test_too_many_requests_normal(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "POST",
            self.webhook_1.url,
            status_code=429,
            json={
                "global": False,
                "message": "You are being rate limited.",
                "retry_after": 2000,
            },
            headers={
                "x-ratelimit-remaining": "5",
                "x-ratelimit-reset-after": "60",
                "Retry-After": "2000",
            },
        )
        # when/then
        try:
            self.webhook_1.send_message_to_webhook(self.message)
        except Exception as ex:
            self.assertIsInstance(ex, WebhookTooManyRequests)
            self.assertEqual(ex.retry_after, 2002)
        else:
            self.fail("Did not raise excepted exception")

        self.assertAlmostEqual(
            cache.ttl(self.webhook_1._blocked_cache_key()), 2002, delta=5
        )

    def test_too_many_requests_no_retry_value(self, requests_mocker):
        # given
        requests_mocker.register_uri(
            "POST",
            self.webhook_1.url,
            status_code=429,
            headers={
                "x-ratelimit-remaining": "5",
                "x-ratelimit-reset-after": "60",
            },
        )
        # when/then
        try:
            self.webhook_1.send_message_to_webhook(self.message)
        except Exception as ex:
            self.assertIsInstance(ex, WebhookTooManyRequests)
            self.assertEqual(ex.retry_after, WebhookTooManyRequests.DEFAULT_RESET_AFTER)
        else:
            self.fail("Did not raise excepted exception")


class TestEveKillmail(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        killmail = load_killmail(10000001)
        cls.eve_killmail = EveKillmail.objects.create_from_killmail(killmail)

    def test_str(self):
        self.assertEqual(str(self.eve_killmail), "ID:10000001")

    def test_repr(self):
        self.assertEqual(repr(self.eve_killmail), "EveKillmail(id=10000001)")

    def test_entity_ids(self):
        result = self.eve_killmail.entity_ids()
        expected = {
            1011,
            2011,
            3011,
            603,
            30004984,
            1001,
            1002,
            1003,
            2001,
            3001,
            34562,
            2977,
            3756,
            2488,
            500001,
            500004,
        }
        self.assertSetEqual(result, expected)


@patch(MODELS_PATH + ".Webhook.enqueue_message")
class TestTrackerGenerateKillmailMessage(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.tracker = TrackerFactory(name="My Tracker", webhook=self.webhook_1)

    def test_should_generate_message(self, mock_enqueue_message):
        # given
        self.tracker.origin_solar_system_id = 30003067
        self.tracker.save()
        svipul = EveType.objects.get(name="Svipul")
        self.tracker.require_attackers_ship_types.add(svipul)
        self.tracker.require_attackers_ship_types.add(
            EveType.objects.get(name="Gnosis")
        )
        killmail = load_killmail(10000101)
        killmail_json = Killmail.from_json(killmail.asjson())
        # when
        self.tracker.generate_killmail_message(killmail_json)
        # then
        _, kwargs = mock_enqueue_message.call_args
        content = kwargs["content"]
        self.assertIn("My Tracker", content)
        embed = kwargs["embeds"][0]
        self.assertEqual(embed.title, "Haras | Svipul | Killmail")
        self.assertEqual(embed.thumbnail.url, svipul.icon_url(size=128))
        html = markdown(embed.description)
        description = "".join(
            BeautifulSoup(html, features="html.parser").findAll(text=True)
        )
        lines = description.splitlines()
        self.assertEqual(
            (
                "Lex Luthor (LexCorp) lost their Svipul in Haras (The Bleak Lands) "
                "worth 10.00k ISK."
            ),
            lines[0],
        )
        self.assertEqual(
            "Final blow by Bruce Wayne (Wayne Technologies) in a Svipul.", lines[1]
        )



import datetime as dt

import factory
import factory.fuzzy

from django.utils.timezone import now

from killtracker.app_settings import KILLTRACKER_KILLMAIL_MAX_AGE_FOR_TRACKER
from killtracker.core.killmails import (
    Killmail,
    KillmailAttacker,
    KillmailPosition,
    KillmailVictim,
    KillmailZkb,
    _KillmailCharacter,
)
from killtracker.models import Tracker, Webhook

from .load_eveuniverse import eveuniverse_testdata

_SOLAR_SYSTEM_IDS = {obj["id"] for obj in eveuniverse_testdata["EveSolarSystem"]}


class KillmailCharacterFactory(factory.Factory):
    class Meta:
        model = _KillmailCharacter

    character_id = factory.Sequence(lambda n: 90_000_001 + n)
    corporation_id = factory.Sequence(lambda n: 98_000_001 + n)
    alliance_id = factory.Sequence(lambda n: 99_000_001 + n)
    ship_type_id = 3756  # Gnosis


class KillmailVictimFactory(KillmailCharacterFactory):
    class Meta:
        model = KillmailVictim

    damage_taken = factory.fuzzy.FuzzyInteger(1_000_000)


class KillmailAttackerFactory(KillmailCharacterFactory):
    class Meta:
        model = KillmailAttacker

    damage_done = factory.fuzzy.FuzzyInteger(1_000_000)
    security_status = factory.fuzzy.FuzzyFloat(-10.0, 5)
    weapon_type_id = 2977


class KillmailPositionFactory(factory.Factory):
    class Meta:
        model = KillmailPosition

    x = factory.fuzzy.FuzzyFloat(-10_000, 10_000)
    y = factory.fuzzy.FuzzyFloat(-10_000, 10_000)
    z = factory.fuzzy.FuzzyFloat(-10_000, 10_000)


class KillmailZkbFactory(factory.Factory):
    class Meta:
        model = KillmailZkb

    location_id = factory.Sequence(lambda n: n + 60_000_000)
    hash = factory.fuzzy.FuzzyText()
    fitted_value = factory.fuzzy.FuzzyFloat(10_000, 100_000_000)
    total_value = factory.LazyAttribute(lambda o: o.fitted_value)
    points = factory.fuzzy.FuzzyInteger(1000)
    is_npc = False
    is_solo = False
    is_awox = False


class KillmailFactory(factory.Factory):
    class Meta:
        model = Killmail

    class Params:
        # max age of a killmail in seconds
        max_age = KILLTRACKER_KILLMAIL_MAX_AGE_FOR_TRACKER

    id = factory.Sequence(lambda n: n + 1800000000001)
    victim = factory.SubFactory(KillmailVictimFactory)
    position = factory.SubFactory(KillmailPositionFactory)
    zkb = factory.SubFactory(KillmailZkbFactory)
    solar_system_id = factory.fuzzy.FuzzyChoice(_SOLAR_SYSTEM_IDS)

    @factory.lazy_attribute
    def time(self):
        return factory.fuzzy.FuzzyDateTime(
            now() - dt.timedelta(seconds=self.max_age - 5)
        ).fuzz()

    @factory.lazy_attribute
    def attackers(self):
        amount = factory.fuzzy.FuzzyInteger(1, 10).fuzz()
        my_attackers = [KillmailAttackerFactory() for _ in range(amount)]
        my_attackers[0].is_final_blow = True
        return my_attackers


class WebhookFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Webhook
        django_get_or_create = ("name",)

    name = factory.Faker("name")
    url = factory.Faker("uri")


class TrackerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Tracker
        django_get_or_create = ("name",)

    name = factory.Faker("name")
    webhook = factory.SubFactory(WebhookFactory)



"""
from django.core.exceptions import ValidationError
from django.contrib.admin.sites import AdminSite

from ..admin import TrackerAdmin
from ..models import Tracker, Webhook
from app_utils.testing import NoSocketsTestCase


class TestTrackerAdmin(NoSocketsTestCase):
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = TrackerAdmin(model=Tracker, admin_site=AdminSite())
        cls.webhook_1 = Webhook.objects.create(
            name="Webhook 1", url="http://www.example.com/webhook_1", is_enabled=True
        )

    def test_clean_no_issue(self):
        tracker = Tracker(name="Test", webhook=self.webhook_1)
        self.modeladmin.form.clean()

    def test_clean_need_origin_for_max_jumps(self):
        tracker = Tracker(name="Test", webhook=self.webhook_1, require_max_jumps=10)
        with self.assertRaises(ValidationError):
            tracker.clean()

    def test_clean_need_origin_for_max_distance(self):
        tracker = Tracker(name="Test", webhook=self.webhook_1, require_max_distance=10)
        with self.assertRaises(ValidationError):
            tracker.clean()
"""



from datetime import timedelta
from typing import Dict, Tuple

from django.db import models, transaction
from django.utils.timezone import now
from eveuniverse.models import EveEntity

from allianceauth.services.hooks import get_extension_logger
from app_utils.caching import ObjectCacheMixin
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS
from .core.killmails import Killmail, _KillmailCharacter

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class EveTypePlusManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        return qs.select_related("eve_group")


class EveKillmailQuerySet(models.QuerySet):
    """Custom queryset for EveKillmail"""

    def load_entities(self) -> int:
        """Load all unknown entities for killmails in this QuerySet.

        Return count of updated entities.
        """
        entity_ids = set()
        for killmail in self:
            entity_ids |= killmail.entity_ids()
        return EveEntity.objects.filter(id__in=entity_ids, name="").update_from_esi()


class EveKillmailBaseManager(models.Manager):
    def delete_stale(self) -> Tuple[int, Dict[str, int]]:
        """deletes all stale killmail"""
        if KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS > 0:
            deadline = now() - timedelta(days=KILLTRACKER_PURGE_KILLMAILS_AFTER_DAYS)
            return self.filter(time__lt=deadline).delete()

    def create_from_killmail(
        self, killmail: Killmail, resolve_ids=True
    ) -> models.Model:
        """create a new EveKillmail from a Killmail object and returns it

        Args:
        - resolve_ids: When set to False will not resolve EveEntity IDs

        """
        from .models import EveKillmailAttacker

        params = {
            "id": killmail.id,
            "time": killmail.time,
            "damage_taken": killmail.victim.damage_taken,
            "position_x": killmail.position.x,
            "position_y": killmail.position.y,
            "position_z": killmail.position.z,
        }
        victim = self._create_args_for_entities(killmail.victim)
        params.update(victim)
        if killmail.solar_system_id:
            params["solar_system"], _ = EveEntity.objects.get_or_create(
                id=killmail.solar_system_id
            )
        if killmail.zkb:
            zkb = killmail.zkb.asdict()
            zkb["zkb_points"] = zkb.pop("points")
            params.update(zkb)
        eve_killmail = self.create(**params)
        if killmail.attackers:
            attacker_objs = []
            for attacker in killmail.attackers:
                params = {
                    **{
                        "killmail": eve_killmail,
                        "damage_done": attacker.damage_done,
                        "security_status": attacker.security_status,
                        "is_final_blow": attacker.is_final_blow,
                    },
                    **self._create_args_for_entities(attacker),
                }
                attacker_objs.append(EveKillmailAttacker(**params))
            EveKillmailAttacker.objects.bulk_create(attacker_objs)
        if resolve_ids:
            eve_killmail.load_entities()
        return eve_killmail

    @staticmethod
    def _create_args_for_entities(killmail_character: _KillmailCharacter) -> dict:
        args = dict()
        for prop_name in killmail_character.ENTITY_PROPS:
            entity_id = getattr(killmail_character, prop_name)
            if entity_id:
                field = prop_name.replace("_id", "")
                args[field], _ = EveEntity.objects.get_or_create(id=entity_id)
        return args

    def update_or_create_from_killmail(
        self, killmail: Killmail
    ) -> Tuple[models.Model, bool]:
        with transaction.atomic():
            try:
                self.get(id=killmail.id).delete()
                created = False
            except self.model.DoesNotExist:
                created = True
            obj = self.create_from_killmail(killmail, resolve_ids=False)
        obj.load_entities()
        return obj, created


EveKillmailManager = EveKillmailBaseManager.from_queryset(EveKillmailQuerySet)


class TrackerManager(ObjectCacheMixin, models.Manager):
    pass


class WebhookManager(ObjectCacheMixin, models.Manager):
    pass



import json
from datetime import timedelta
from unittest.mock import Mock, patch

from bravado.exception import HTTPNotFound

from django.test import TestCase
from django.utils.timezone import now
from eveuniverse.models import (
    EveConstellation,
    EveGroup,
    EveRegion,
    EveSolarSystem,
    EveType,
)

from allianceauth.eveonline.models import EveAllianceInfo, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.esi_testing import BravadoOperationStub
from app_utils.testing import NoSocketsTestCase, add_character_to_user_2

from ..core.killmails import EntityCount, Killmail
from ..models import Tracker
from .testdata.factories import (
    KillmailAttackerFactory,
    KillmailFactory,
    KillmailVictimFactory,
    TrackerFactory,
)
from .testdata.helpers import LoadTestDataMixin, load_killmail

MODULE_PATH = "killtracker.models"


def esi_get_route_origin_destination(origin, destination, **kwargs) -> list:
    routes = {
        30003067: {
            30003087: [
                30003067,
                30003068,
                30003069,
                30003070,
                30003071,
                30003091,
                30003086,
                30003087,
            ],
            30003070: [30003067, 30003068, 30003069, 30003070],
            30003067: [30003067],
        },
    }
    if origin in routes and destination in routes[origin]:
        return BravadoOperationStub(routes[origin][destination])
    else:
        raise HTTPNotFound(Mock(**{"response.status_code": 404}))


class TestTrackerCalculate(LoadTestDataMixin, NoSocketsTestCase):
    @classmethod
    def _matching_killmail_ids(cls, tracker: Tracker, killmail_ids: set) -> set:
        return {
            killmail.id for killmail in cls._matching_killmails(tracker, killmail_ids)
        }

    @staticmethod
    def _matching_killmails(tracker: Tracker, killmail_ids: set) -> list:
        results = list()
        for killmail_id in killmail_ids:
            killmail = load_killmail(killmail_id)
            new_killmail = tracker.process_killmail(killmail)
            if new_killmail:
                results.append(new_killmail)
        return results

    def test_can_match_all(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000002, 10000003, 10000004, 10000005}
        self.assertSetEqual(results, expected)

    @patch(MODULE_PATH + ".KILLTRACKER_KILLMAIL_MAX_AGE_FOR_TRACKER", 60)
    def test_excludes_older_killmails(self):
        tracker = TrackerFactory(
            name="Test",
            webhook=self.webhook_1,
        )
        killmail_1 = load_killmail(10000001)
        killmail_2 = load_killmail(10000002)
        killmail_2.time = now() - timedelta(hours=1, seconds=1)
        results = set()
        for killmail in [killmail_1, killmail_2]:
            if tracker.process_killmail(killmail):
                results.add(killmail.id)

        expected = {10000001}
        self.assertSetEqual(results, expected)

    def test_can_process_killmail_without_solar_system(self):
        tracker = TrackerFactory(exclude_high_sec=True, webhook=self.webhook_1)
        self.assertIsNotNone(tracker.process_killmail(load_killmail(10000402)))

    def test_can_filter_high_sec_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(exclude_high_sec=True, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000003, 10000004}
        self.assertSetEqual(results, expected)

    def test_can_filter_low_sec_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(exclude_low_sec=True, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000002, 10000003, 10000004}
        self.assertSetEqual(results, expected)

    def test_can_filter_null_sec_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(exclude_null_sec=True, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000002, 10000004}
        self.assertSetEqual(results, expected)

    def test_can_filter_w_space_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(exclude_w_space=True, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000002, 10000003}
        self.assertSetEqual(results, expected)

    def test_can_filter_min_attackers(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(require_min_attackers=3, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001}
        self.assertSetEqual(results, expected)

    def test_can_filter_max_attackers(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(require_max_attackers=2, webhook=self.webhook_1)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000002, 10000003, 10000004}
        self.assertSetEqual(results, expected)

    @patch("eveuniverse.models.esi")
    def test_can_filter_max_jumps(self, mock_esi):
        mock_esi.client.Routes.get_route_origin_destination.side_effect = (
            esi_get_route_origin_destination
        )

        killmail_ids = {10000101, 10000102, 10000103}
        tracker = TrackerFactory(
            origin_solar_system_id=30003067,
            require_max_jumps=3,
            webhook=self.webhook_1,
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000102, 10000103}
        self.assertSetEqual(results, expected)

    @patch("eveuniverse.models.esi")
    def test_can_filter_max_distance(self, mock_esi):
        mock_esi.client.Routes.get_route_origin_destination.side_effect = (
            esi_get_route_origin_destination
        )

        killmail_ids = {10000101, 10000102, 10000103}
        tracker = TrackerFactory(
            origin_solar_system_id=30003067,
            require_max_distance=2,
            webhook=self.webhook_1,
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000102, 10000103}
        self.assertSetEqual(results, expected)

    def test_can_filter_attacker_alliance(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_attacker_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_filter_attacker_corporation(self):
        killmail_ids = {10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2001)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_required_attacker_alliance(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3011)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_required_attacker_corporation(self):
        killmail_ids = {10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2011)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_required_victim_alliances(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_victim_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_required_victim_corporation(self):
        killmail_ids = {10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_victim_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2001)
        )
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000005}
        self.assertSetEqual(results, expected)

    def test_can_filter_nullsec_and_attacker_alliance(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(exclude_null_sec=True, webhook=self.webhook_1)
        excluded_alliance = EveAllianceInfo.objects.get(alliance_id=3001)
        tracker.require_attacker_alliances.add(excluded_alliance)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000002, 10000004}
        self.assertSetEqual(results, expected)

    def test_can_require_region(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_regions.add(EveRegion.objects.get(id=10000014))
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000003}
        self.assertSetEqual(results, expected)

    def test_can_require_constellation(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_constellations.add(EveConstellation.objects.get(id=20000169))
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000003}
        self.assertSetEqual(results, expected)

    def test_can_require_solar_system(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_solar_systems.add(EveSolarSystem.objects.get(id=30001161))
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000003}
        self.assertSetEqual(results, expected)

    def test_can_require_attackers_ship_groups(self):
        killmail_ids = {10000101, 10000201}
        tracker = TrackerFactory(webhook=self.webhook_1)
        frigate = EveGroup.objects.get(id=25)
        td3s = EveGroup.objects.get(id=1305)
        tracker.require_attackers_ship_groups.add(frigate)
        tracker.require_attackers_ship_groups.add(td3s)
        results = self._matching_killmails(tracker, killmail_ids)
        self.assertEqual(len(results), 1)
        killmail = results[0]
        self.assertEqual(killmail.id, 10000101)
        self.assertListEqual(killmail.tracker_info.matching_ship_type_ids, [34562])

    def test_can_require_victim_ship_group(self):
        killmail_ids = {10000101, 10000201}
        tracker = TrackerFactory(webhook=self.webhook_1)
        td3s = EveGroup.objects.get(id=1305)
        tracker.require_victim_ship_groups.add(td3s)
        results = self._matching_killmails(tracker, killmail_ids)
        self.assertEqual(len(results), 1)
        killmail = results[0]
        self.assertEqual(killmail.id, 10000101)
        self.assertListEqual(killmail.tracker_info.matching_ship_type_ids, [34562])

    def test_can_require_victim_ship_types(self):
        killmail_ids = {10000101, 10000201}
        tracker = TrackerFactory(webhook=self.webhook_1)
        svipul = EveType.objects.get(id=34562)
        tracker.require_victim_ship_types.add(svipul)
        results = self._matching_killmails(tracker, killmail_ids)
        self.assertEqual(len(results), 1)
        killmail = results[0]
        self.assertEqual(killmail.id, 10000101)
        self.assertListEqual(killmail.tracker_info.matching_ship_type_ids, [34562])

    def test_can_require_attackers_ship_types(self):
        """
        when filtering for attackers with ship groups of Frigate, TD3
        then tracker finds killmail that has attacker with TD3 and no attacker with frigate
        and ignores killmail that attackers with neither
        """
        killmail_ids = {10000101, 10000201}
        tracker = TrackerFactory(webhook=self.webhook_1)
        svipul = EveType.objects.get(id=34562)
        tracker.require_attackers_ship_types.add(svipul)
        results = self._matching_killmails(tracker, killmail_ids)
        self.assertEqual(len(results), 1)
        killmail = results[0]
        self.assertEqual(killmail.id, 10000101)
        self.assertListEqual(killmail.tracker_info.matching_ship_type_ids, [34562])

    def test_can_exclude_npc_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005, 10000301}
        tracker = TrackerFactory(webhook=self.webhook_1, exclude_npc_kills=True)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000001, 10000002, 10000003, 10000004, 10000005}
        self.assertSetEqual(results, expected)

    def test_can_require_npc_kills(self):
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005, 10000301}
        tracker = TrackerFactory(webhook=self.webhook_1, require_npc_kills=True)
        results = self._matching_killmail_ids(tracker, killmail_ids)
        expected = {10000301}
        self.assertSetEqual(results, expected)

    def test_should_apply_require_attackers_states(self):
        # given
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_states.add(self.state_member)
        user = AuthUtils.create_member("Lex Luther")
        add_character_to_user_2(user, 1011, "Lex Luthor", 2011, "LexCorp")
        # when
        results = self._matching_killmail_ids(tracker, killmail_ids)
        # then
        self.assertSetEqual(results, {10000005})

    def test_should_apply_exclude_attacker_states(self):
        # given
        killmail_ids = {10000001, 10000002, 10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_attacker_states.add(self.state_member)
        user = AuthUtils.create_member("Lex Luther")
        add_character_to_user_2(user, 1011, "Lex Luthor", 2011, "LexCorp")
        # when
        results = self._matching_killmail_ids(tracker, killmail_ids)
        # then
        self.assertSetEqual(results, {10000001, 10000002, 10000003, 10000004})

    def test_should_apply_require_victim_states(self):
        # given
        killmail_ids = {10000003, 10000004, 10000005}
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_victim_states.add(self.state_member)
        user = AuthUtils.create_member("Lex Luther")
        add_character_to_user_2(user, 1011, "Lex Luthor", 2011, "LexCorp")
        # when
        results = self._matching_killmail_ids(tracker, killmail_ids)
        # then
        self.assertSetEqual(results, {10000003, 10000004})


class TestTrackerCalculate2(LoadTestDataMixin, NoSocketsTestCase):
    def test_should_accept_with_require_attacker_alliances(self):
        attacker = KillmailAttackerFactory(alliance_id=3011)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_deny_with_require_attacker_alliances(self):
        attacker = KillmailAttackerFactory(alliance_id=3010)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_accept_with_require_attacker_corporations(self):
        attacker = KillmailAttackerFactory(corporation_id=2011)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_deny_with_require_attacker_corporations(self):
        attacker = KillmailAttackerFactory(alliance_id=2010)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.require_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_deny_with_exclude_victim_corporation(self):
        victim = KillmailVictimFactory(corporation_id=2001)
        killmail = KillmailFactory(victim=victim)
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_victim_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2001)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_accept_with_exclude_victim_corporation(self):
        victim = KillmailVictimFactory(corporation_id=2002)
        killmail = KillmailFactory(victim=victim)
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_victim_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2001)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_deny_with_exclude_victim_alliance(self):
        victim = KillmailVictimFactory(alliance_id=3001)
        killmail = KillmailFactory(victim=victim)
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_victim_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_accept_with_exclude_victim_alliance(self):
        victim = KillmailVictimFactory(alliance_id=3002)
        killmail = KillmailFactory(victim=victim)
        tracker = TrackerFactory(webhook=self.webhook_1)
        tracker.exclude_victim_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3001)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_accept_when_attacker_alliance_has_final_blow(self):
        attacker = KillmailAttackerFactory(alliance_id=3011, is_final_blow=True)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(
            require_attacker_organizations_final_blow=True, webhook=self.webhook_1
        )
        tracker.require_attacker_alliances.add(
            EveAllianceInfo.objects.get(alliance_id=3011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_accept_when_attacker_corporation_has_final_blow(self):
        attacker = KillmailAttackerFactory(corporation_id=2011, is_final_blow=True)
        killmail = KillmailFactory(attackers=[attacker])
        tracker = TrackerFactory(
            require_attacker_organizations_final_blow=True, webhook=self.webhook_1
        )
        tracker.require_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNotNone(result)

    def test_should_deny_when_attacker_corporation_has_not_has_final_blow(self):
        attacker_1 = KillmailAttackerFactory(corporation_id=2011, is_final_blow=False)
        attacker_2 = KillmailAttackerFactory(corporation_id=2001, is_final_blow=True)
        killmail = KillmailFactory(attackers=[attacker_1, attacker_2])
        tracker = TrackerFactory(
            require_attacker_organizations_final_blow=True, webhook=self.webhook_1
        )
        tracker.require_attacker_corporations.add(
            EveCorporationInfo.objects.get(corporation_id=2011)
        )
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_deny_when_value_is_below_minimum(self):
        killmail = KillmailFactory(zkb__total_value=50_000_000)
        tracker = TrackerFactory(require_min_value=51, webhook=self.webhook_1)
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)

    def test_should_threat_no_value_as_zero(self):
        killmail = KillmailFactory(zkb__total_value=None)
        tracker = TrackerFactory(require_min_value=51, webhook=self.webhook_1)
        # when
        result = tracker.process_killmail(killmail)
        # then
        self.assertIsNone(result)


class TestTrackerCalculateTrackerInfo(LoadTestDataMixin, NoSocketsTestCase):
    def setUp(self) -> None:
        self.tracker = TrackerFactory(webhook=self.webhook_1)

    @patch("eveuniverse.models.esi")
    def test_basics(self, mock_esi):
        # given
        mock_esi.client.Routes.get_route_origin_destination.side_effect = (
            esi_get_route_origin_destination
        )
        self.tracker.origin_solar_system_id = 30003067
        self.tracker.save()
        # when
        killmail = self.tracker.process_killmail(load_killmail(10000101))
        # then
        self.assertTrue(killmail.tracker_info)
        self.assertEqual(killmail.tracker_info.tracker_pk, self.tracker.pk)
        self.assertEqual(killmail.tracker_info.jumps, 7)
        self.assertAlmostEqual(killmail.tracker_info.distance, 5.85, delta=0.01)
        self.assertEqual(
            killmail.tracker_info.main_org,
            EntityCount(id=3001, category=EntityCount.CATEGORY_ALLIANCE, count=3),
        )
        self.assertEqual(
            killmail.tracker_info.main_ship_group,
            EntityCount(
                id=419,
                category=EntityCount.CATEGORY_INVENTORY_GROUP,
                name="Combat Battlecruiser",
                count=2,
            ),
        )

    def test_main_org_corporation_is_main(self):
        killmail = self.tracker.process_killmail(load_killmail(10000403))
        self.assertEqual(
            killmail.tracker_info.main_org,
            EntityCount(id=2001, category=EntityCount.CATEGORY_CORPORATION, count=2),
        )

    def test_main_org_prioritize_alliance_over_corporation(self):
        killmail = self.tracker.process_killmail(load_killmail(10000401))
        self.assertEqual(
            killmail.tracker_info.main_org,
            EntityCount(id=3001, category=EntityCount.CATEGORY_ALLIANCE, count=2),
        )

    def test_main_org_is_none_if_only_one_attacker(self):
        killmail = self.tracker.process_killmail(load_killmail(10000005))
        self.assertIsNone(killmail.tracker_info.main_org)

    def test_main_org_is_none_if_faction(self):
        killmail = self.tracker.process_killmail(load_killmail(10000302))
        self.assertIsNone(killmail.tracker_info.main_org)

    def test_main_ship_group_above_threshold(self):
        killmail = self.tracker.process_killmail(load_killmail(10006001))
        self.assertEqual(
            killmail.tracker_info.main_ship_group,
            EntityCount(
                id=419, category="inventory_group", name="Combat Battlecruiser", count=2
            ),
        )

    def test_main_ship_group_return_none_if_below_threshold(self):
        killmail = self.tracker.process_killmail(load_killmail(10006002))
        self.assertIsNone(killmail.tracker_info.main_ship_group)

    def test_main_org_above_threshold(self):
        killmail = self.tracker.process_killmail(load_killmail(10006003))
        self.assertEqual(
            killmail.tracker_info.main_org,
            EntityCount(id=2001, category="corporation", count=2),
        )

    def test_main_org_return_none_if_below_threshold(self):
        killmail = self.tracker.process_killmail(load_killmail(10006004))
        self.assertIsNone(killmail.tracker_info.main_org)


class TestTrackerEnqueueKillmail(LoadTestDataMixin, TestCase):
    def setUp(self) -> None:
        self.tracker = TrackerFactory(name="My Tracker", webhook=self.webhook_1)
        self.webhook_1.main_queue.clear()

    @patch(MODULE_PATH + ".KILLTRACKER_WEBHOOK_SET_AVATAR", True)
    @patch("eveuniverse.models.esi")
    def test_normal(self, mock_esi):
        mock_esi.client.Routes.get_route_origin_destination.side_effect = (
            esi_get_route_origin_destination
        )
        self.tracker.origin_solar_system_id = 30003067
        self.tracker.save()
        svipul = EveType.objects.get(id=34562)
        self.tracker.require_attackers_ship_types.add(svipul)
        gnosis = EveType.objects.get(id=3756)
        self.tracker.require_attackers_ship_types.add(gnosis)
        killmail = self.tracker.process_killmail(load_killmail(10000101))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())

        self.assertEqual(message["username"], "Killtracker")
        self.assertIsNotNone(message["avatar_url"])
        self.assertIn("My Tracker", message["content"])
        embed = message["embeds"][0]
        self.assertIn("| Killmail", embed["title"])
        self.assertIn("Combat Battlecruiser", embed["description"])
        self.assertIn("Tracked ship types", embed["description"])

    @patch(MODULE_PATH + ".KILLTRACKER_WEBHOOK_SET_AVATAR", False)
    @patch("eveuniverse.models.esi")
    def test_disabled_avatar(self, mock_esi):
        mock_esi.client.Routes.get_route_origin_destination.side_effect = (
            esi_get_route_origin_destination
        )
        self.tracker.origin_solar_system_id = 30003067
        self.tracker.save()
        svipul = EveType.objects.get(id=34562)
        self.tracker.require_attackers_ship_types.add(svipul)
        gnosis = EveType.objects.get(id=3756)
        self.tracker.require_attackers_ship_types.add(gnosis)
        killmail = self.tracker.process_killmail(load_killmail(10000101))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertNotIn("username", message)
        self.assertNotIn("avatar_url", message)
        self.assertIn("My Tracker", message["content"])

    def test_send_as_fleetkill(self):
        self.tracker.identify_fleets = True
        self.tracker.save()
        killmail = self.tracker.process_killmail(load_killmail(10000101))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertIn("| Fleetkill", message["embeds"][0]["title"])

    def test_can_add_intro_text(self):
        killmail = self.tracker.process_killmail(load_killmail(10000101))

        self.tracker.generate_killmail_message(killmail, intro_text="Intro Text")

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertIn("Intro Text", message["content"])

    def test_without_tracker_info(self):
        killmail = load_killmail(10000001)

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)

    def test_can_ping_everybody(self):
        tracker = TrackerFactory(
            webhook=self.webhook_1, ping_type=Tracker.ChannelPingType.EVERYBODY
        )

        killmail = tracker.process_killmail(load_killmail(10000001))

        tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertIn("@everybody", message["content"])

    def test_can_ping_here(self):
        self.tracker.ping_type = Tracker.ChannelPingType.HERE
        self.tracker.save()

        killmail = self.tracker.process_killmail(load_killmail(10000001))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertIn("@here", message["content"])

    def test_can_ping_nobody(self):
        self.tracker.ping_type = Tracker.ChannelPingType.NONE
        self.tracker.save()
        killmail = self.tracker.process_killmail(load_killmail(10000001))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertNotIn("@everybody", message["content"])
        self.assertNotIn("@here", message["content"])

    def test_can_disable_posting_name(self):
        self.tracker.s_posting_name = False
        self.tracker.save()
        killmail = self.tracker.process_killmail(load_killmail(10000001))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)
        message = json.loads(self.webhook_1.main_queue.dequeue())
        self.assertNotIn("Ping Nobody", message["content"])

    def test_can_send_npc_killmail(self):
        killmail = self.tracker.process_killmail(load_killmail(10000301))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)

    def test_can_handle_victim_without_character(self):
        killmail = self.tracker.process_killmail(load_killmail(10000501))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)

    def test_can_handle_victim_without_corporation(self):
        killmail = self.tracker.process_killmail(load_killmail(10000502))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)

    def test_can_handle_final_attacker_with_no_character(self):
        killmail = self.tracker.process_killmail(load_killmail(10000503))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)

    def test_can_handle_matching_type_ids(self):
        svipul = EveType.objects.get(id=34562)
        self.tracker.require_attackers_ship_types.add(svipul)
        killmail = self.tracker.process_killmail(load_killmail(10000001))

        self.tracker.generate_killmail_message(Killmail.from_json(killmail.asjson()))

        self.assertEqual(self.webhook_1.main_queue.size(), 1)



# Generated by Django 4.0.7 on 2022-09-03 18:33

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("killtracker", "0008_copy_data_to_new_structure"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="alliance",
        ),
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="character",
        ),
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="corporation",
        ),
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="faction",
        ),
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="killmail",
        ),
        migrations.RemoveField(
            model_name="evekillmailvictim",
            name="ship_type",
        ),
        migrations.RemoveField(
            model_name="evekillmailzkb",
            name="killmail",
        ),
        migrations.DeleteModel(
            name="EveKillmailPosition",
        ),
        migrations.DeleteModel(
            name="EveKillmailVictim",
        ),
        migrations.DeleteModel(
            name="EveKillmailZkb",
        ),
    ]



# Generated by Django 4.0.7 on 2022-10-12 11:33
# Manually adopted to replace a squashed migration and be compatible with AA 2.x

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [
        ("killtracker", "0001_squashed_all"),
        ("killtracker", "0002_fix_webhook_notes_field"),
        ("killtracker", "0003_add_state_clauses"),
        ("killtracker", "0004_django4_update"),
        ("killtracker", "0005_add_final_blow_clause_and_more"),
        ("killtracker", "0006_evetypeplus"),
        ("killtracker", "0007_restructure_killsmails"),
        ("killtracker", "0008_copy_data_to_new_structure"),
        ("killtracker", "0009_remove_old_models"),
    ]

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        ("authentication", "0019_merge_20211026_0919"),
        ("eveuniverse", "0007_evetype_description"),
        ("eveonline", "0015_factions"),
    ]

    operations = [
        migrations.CreateModel(
            name="EveKillmail",
            fields=[
                ("id", models.BigIntegerField(primary_key=True, serialize=False)),
                (
                    "time",
                    models.DateTimeField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "damage_taken",
                    models.BigIntegerField(blank=True, default=None, null=True),
                ),
                ("position_x", models.FloatField(blank=True, default=None, null=True)),
                ("position_y", models.FloatField(blank=True, default=None, null=True)),
                ("position_z", models.FloatField(blank=True, default=None, null=True)),
                (
                    "location_id",
                    models.PositiveIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                ("hash", models.CharField(blank=True, default="", max_length=64)),
                (
                    "fitted_value",
                    models.FloatField(blank=True, default=None, null=True),
                ),
                (
                    "total_value",
                    models.FloatField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "zkb_points",
                    models.PositiveIntegerField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_npc",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_solo",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "is_awox",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "ship_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="short name to identify this webhook",
                        max_length=64,
                        unique=True,
                    ),
                ),
                (
                    "webhook_type",
                    models.IntegerField(
                        choices=[(1, "Discord Webhook")],
                        default=1,
                        help_text="type of this webhook",
                    ),
                ),
                (
                    "url",
                    models.CharField(
                        help_text="URL of this webhook, e.g. https://discordapp.com/api/webhooks/123456/abcdef",
                        max_length=255,
                        unique=True,
                    ),
                ),
                (
                    "notes",
                    models.TextField(
                        blank=True,
                        help_text="you can add notes about this webhook here if you want",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="whether notifications are currently sent to this webhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveTypePlus",
            fields=[],
            options={
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("eveuniverse.evetype",),
        ),
        migrations.CreateModel(
            name="Tracker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Name to identify tracker. Will be shown on alerts posts.",
                        max_length=100,
                        unique=True,
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Brief description what this tracker is for. Will not be shown on alerts.",
                    ),
                ),
                (
                    "color",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="Optional color for embed on Discord - #000000 / black means no color selected.",
                        max_length=7,
                    ),
                ),
                (
                    "require_max_jumps",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require all killmails to be max x jumps away from origin solar system.",
                        null=True,
                    ),
                ),
                (
                    "require_max_distance",
                    models.FloatField(
                        blank=True,
                        default=None,
                        help_text="Require all killmails to be max x LY away from origin solar system.",
                        null=True,
                    ),
                ),
                (
                    "require_attacker_organizations_final_blow",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="Only include killmails where at least one of the specified <b>required attacker corporations</b> or <b>required attacker alliances</b> has the final blow.",
                    ),
                ),
                (
                    "identify_fleets",
                    models.BooleanField(
                        default=False,
                        help_text="When true: kills are interpreted and shown as fleet kills.",
                    ),
                ),
                (
                    "exclude_blue_attackers",
                    models.BooleanField(
                        default=False,
                        help_text="Exclude killmails with blue attackers.",
                    ),
                ),
                (
                    "require_blue_victim",
                    models.BooleanField(
                        default=False,
                        help_text="Only include killmails where the victim has standing with our group.",
                    ),
                ),
                (
                    "require_min_attackers",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmails to have at least given number of attackers.",
                        null=True,
                    ),
                ),
                (
                    "require_max_attackers",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmails to have no more than max number of attackers.",
                        null=True,
                    ),
                ),
                (
                    "exclude_high_sec",
                    models.BooleanField(
                        default=False,
                        help_text="Exclude killmails from high sec. Also exclude high sec systems in route finder for jumps from origin.",
                    ),
                ),
                (
                    "exclude_low_sec",
                    models.BooleanField(
                        default=False, help_text="Exclude killmails from low sec."
                    ),
                ),
                (
                    "exclude_null_sec",
                    models.BooleanField(
                        default=False, help_text="Exclude killmails from null sec."
                    ),
                ),
                (
                    "exclude_w_space",
                    models.BooleanField(
                        default=False, help_text="Exclude killmails from WH space."
                    ),
                ),
                (
                    "require_min_value",
                    models.PositiveIntegerField(
                        blank=True,
                        default=None,
                        help_text="Require killmail's value to be greater or equal to the given value in M ISK.",
                        null=True,
                    ),
                ),
                (
                    "exclude_npc_kills",
                    models.BooleanField(default=False, help_text="Exclude npc kills."),
                ),
                (
                    "require_npc_kills",
                    models.BooleanField(
                        default=False,
                        help_text="Only include killmails that are npc kills.",
                    ),
                ),
                (
                    "ping_type",
                    models.CharField(
                        choices=[
                            ("PN", "(none)"),
                            ("PH", "@here"),
                            ("PE", "@everybody"),
                        ],
                        default="PN",
                        help_text="Option to ping every member of the channel.",
                        max_length=2,
                        verbose_name="channel pings",
                    ),
                ),
                (
                    "is_posting_name",
                    models.BooleanField(
                        default=True,
                        help_text="Whether posted messages include the tracker's name.",
                    ),
                ),
                (
                    "is_enabled",
                    models.BooleanField(
                        db_index=True,
                        default=True,
                        help_text="Toogle for activating or deactivating a tracker.",
                    ),
                ),
                (
                    "exclude_attacker_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Exclude killmails with attackers from one of these alliances. ",
                        related_name="+",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "exclude_attacker_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Exclude killmails with attackers from one of these corporations. ",
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "exclude_attacker_states",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Exclude killmails with characters belonging to users with these Auth states. ",
                        related_name="+",
                        to="authentication.state",
                    ),
                ),
                (
                    "exclude_victim_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Exclude killmails where the victim belongs to one of these alliances. ",
                        related_name="+",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "exclude_victim_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Exclude killmails where the victim belongs to one of these corporations. ",
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "origin_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="Solar system to calculate distance and jumps from. When provided distance and jumps will be shown on killmail messages.",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "ping_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Option to ping specific group members. ",
                        related_name="+",
                        to="auth.group",
                        verbose_name="group pings",
                    ),
                ),
                (
                    "require_attacker_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails with attackers from one of these alliances. ",
                        related_name="+",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "require_attacker_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails with attackers from one of these corporations. ",
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "require_attacker_states",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails with characters belonging to users with these Auth states. ",
                        related_name="+",
                        to="authentication.state",
                    ),
                ),
                (
                    "require_attackers_ship_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where at least one attacker is flying one of these ship groups. ",
                        related_name="+",
                        to="eveuniverse.evegroup",
                    ),
                ),
                (
                    "require_attackers_ship_types",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where at least one attacker is flying one of these ship types. ",
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "require_constellations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions. ",
                        related_name="+",
                        to="eveuniverse.eveconstellation",
                    ),
                ),
                (
                    "require_regions",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions. ",
                        related_name="+",
                        to="eveuniverse.everegion",
                    ),
                ),
                (
                    "require_solar_systems",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails that occurred in one of these regions. ",
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                    ),
                ),
                (
                    "require_victim_alliances",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where the victim belongs to one of these alliances. ",
                        related_name="+",
                        to="eveonline.eveallianceinfo",
                    ),
                ),
                (
                    "require_victim_corporations",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where the victim belongs to one of these corporations. ",
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "require_victim_ship_groups",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where victim is flying one of these ship groups. ",
                        related_name="+",
                        to="eveuniverse.evegroup",
                    ),
                ),
                (
                    "require_victim_ship_types",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where victim is flying one of these ship types. ",
                        related_name="+",
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "require_victim_states",
                    models.ManyToManyField(
                        blank=True,
                        default=None,
                        help_text="Only include killmails where the victim characters belong to users with these Auth states. ",
                        related_name="+",
                        to="authentication.state",
                    ),
                ),
                (
                    "webhook",
                    models.ForeignKey(
                        help_text="Webhook URL for a channel on Discord to sent all alerts to.",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="killtracker.webhook",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="EveKillmailAttacker",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "damage_done",
                    models.BigIntegerField(blank=True, default=None, null=True),
                ),
                (
                    "is_final_blow",
                    models.BooleanField(
                        blank=True, db_index=True, default=None, null=True
                    ),
                ),
                (
                    "security_status",
                    models.FloatField(blank=True, default=None, null=True),
                ),
                (
                    "alliance",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "faction",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "killmail",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attackers",
                        to="killtracker.evekillmail",
                    ),
                ),
                (
                    "ship_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
                (
                    "weapon_type",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveuniverse.eveentity",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.test import TestCase
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec("EveFaction", ids=[500001, 500004]),
            ModelSpec(
                "EveType",
                ids=[
                    603,
                    621,
                    638,
                    2488,
                    2977,
                    3756,
                    11379,
                    16238,
                    34562,
                    37483,
                    23320,
                ],
            ),
            ModelSpec(
                "EveSolarSystem", ids=[30001161, 30004976, 30004984, 30045349, 31000005]
            ),
            ModelSpec("EveRegion", ids=[10000038], include_children=True),
        ]
        create_testdata(testdata_spec, test_data_filename())



"""Copy killmail data from old structure to new."""

from django.core.exceptions import ObjectDoesNotExist
from django.db import migrations


def copy_killmails(apps, schema_editor):
    EveKillmail = apps.get_model("killtracker", "EveKillmail")
    print(f"Processing {EveKillmail.objects.count():,} killmails...", end="")
    objs = []
    for killmail in EveKillmail.objects.iterator():
        # victim
        try:
            killmail.character = killmail.victim.character
            killmail.corporation = killmail.victim.corporation
            killmail.alliance = killmail.victim.alliance
            killmail.faction = killmail.victim.faction
            killmail.ship_type = killmail.victim.ship_type
            killmail.damage_taken = killmail.victim.damage_taken
        except ObjectDoesNotExist:
            pass
        # position
        try:
            killmail.position_x = killmail.position.x
            killmail.position_y = killmail.position.y
            killmail.position_z = killmail.position.z
        except ObjectDoesNotExist:
            pass
        # zkb
        try:
            killmail.location_id = killmail.zkb.location_id
            killmail.hash = killmail.zkb.hash
            killmail.fitted_value = killmail.zkb.fitted_value
            killmail.total_value = killmail.zkb.total_value
            killmail.zkb_points = killmail.zkb.points
            killmail.is_npc = killmail.zkb.is_npc
            killmail.is_solo = killmail.zkb.is_solo
            killmail.is_awox = killmail.zkb.is_awox
        except ObjectDoesNotExist:
            pass
        objs.append(killmail)
    EveKillmail.objects.bulk_update(
        objs,
        [
            "character",
            "corporation",
            "alliance",
            "faction",
            "ship_type",
            "damage_taken",
            "position_x",
            "position_y",
            "position_z",
            "location_id",
            "hash",
            "fitted_value",
            "total_value",
            "zkb_points",
            "is_npc",
            "is_solo",
            "is_awox",
        ],
        batch_size=500,
    )


class Migration(migrations.Migration):

    dependencies = [
        ("killtracker", "0007_restructure_killsmails"),
    ]

    operations = [
        migrations.RunPython(copy_killmails, migrations.RunPython.noop),
    ]



import dhooks_lite

from app_utils.testing import NoSocketsTestCase

from killtracker.core import discord_messages

from ..testdata.factories import (
    KillmailAttackerFactory,
    KillmailFactory,
    KillmailVictimFactory,
    TrackerFactory,
)
from ..testdata.helpers import LoadTestDataMixin


class TestCreateEmbed(LoadTestDataMixin, NoSocketsTestCase):
    def test_should_create_normal_embed(self):
        # given
        tracker = TrackerFactory()
        attacker = KillmailAttackerFactory(
            character_id=1011, corporation_id=2011, alliance_id=3011
        )
        victim = KillmailVictimFactory(
            character_id=1001, corporation_id=2001, alliance_id=3001
        )
        killmail = KillmailFactory(victim=victim, attackers=[attacker])
        # when
        embed = discord_messages.create_embed(tracker, killmail)
        # then
        self.assertIsInstance(embed, dhooks_lite.Embed)

    def test_should_create_normal_for_killmail_without_value(self):
        # given
        tracker = TrackerFactory()
        attacker = KillmailAttackerFactory(
            character_id=1011, corporation_id=2011, alliance_id=3011
        )
        victim = KillmailVictimFactory(
            character_id=1001, corporation_id=2001, alliance_id=3001
        )
        killmail = KillmailFactory(
            victim=victim, attackers=[attacker], zkb__total_value=None
        )
        # when
        embed = discord_messages.create_embed(tracker, killmail)
        # then
        self.assertIsInstance(embed, dhooks_lite.Embed)



import json
from unittest.mock import patch

from requests.exceptions import HTTPError

from django.contrib.auth.models import Group
from django.test import TestCase

from app_utils.django import app_labels

from killtracker.core.killmails import Killmail
from killtracker.models import Tracker

from .testdata.factories import TrackerFactory
from .testdata.helpers import LoadTestDataMixin, load_killmail

DISCORD_MESSAGES_PATH = "killtracker.core.discord_messages"

if "discord" in app_labels():

    @patch(DISCORD_MESSAGES_PATH + "._import_discord_user")
    class TestGroupPings(LoadTestDataMixin, TestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            cls.group_1 = Group.objects.create(name="Dummy Group 1")
            cls.group_2 = Group.objects.create(name="Dummy Group 2")

        def setUp(self):
            self.tracker = TrackerFactory(
                webhook=self.webhook_1, exclude_null_sec=True, exclude_w_space=True
            )

        @staticmethod
        def _my_group_to_role(group: Group) -> dict:
            if not isinstance(group, Group):
                raise TypeError("group must be of type Group")

            return {"id": group.pk, "name": group.name}

        def test_can_ping_one_group(self, mock_import_discord_user):
            mock_import_discord_user.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            self.tracker.ping_groups.add(self.group_1)
            killmail = self.tracker.process_killmail(load_killmail(10000101))

            self.tracker.generate_killmail_message(
                Killmail.from_json(killmail.asjson())
            )

            self.assertTrue(
                mock_import_discord_user.return_value.objects.group_to_role.called
            )
            self.assertEqual(self.webhook_1.main_queue.size(), 1)
            message = json.loads(self.webhook_1.main_queue.dequeue())
            self.assertIn(f"<@&{self.group_1.pk}>", message["content"])

        def test_can_ping_multiple_groups(self, mock_import_discord_user):
            mock_import_discord_user.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            self.tracker.ping_groups.add(self.group_1)
            self.tracker.ping_groups.add(self.group_2)

            killmail = self.tracker.process_killmail(load_killmail(10000101))
            self.tracker.generate_killmail_message(
                Killmail.from_json(killmail.asjson())
            )

            self.assertTrue(
                mock_import_discord_user.return_value.objects.group_to_role.called
            )
            self.assertEqual(self.webhook_1.main_queue.size(), 1)
            message = json.loads(self.webhook_1.main_queue.dequeue())
            self.assertIn(f"<@&{self.group_1.pk}>", message["content"])
            self.assertIn(f"<@&{self.group_2.pk}>", message["content"])

        def test_can_combine_with_channel_ping(self, mock_import_discord_user):
            mock_import_discord_user.return_value.objects.group_to_role.side_effect = (
                self._my_group_to_role
            )
            self.tracker.ping_groups.add(self.group_1)
            self.tracker.ping_type = Tracker.ChannelPingType.HERE
            self.tracker.save()

            killmail = self.tracker.process_killmail(load_killmail(10000101))
            self.tracker.generate_killmail_message(
                Killmail.from_json(killmail.asjson())
            )

            self.assertTrue(
                mock_import_discord_user.return_value.objects.group_to_role.called
            )
            self.assertEqual(self.webhook_1.main_queue.size(), 1)
            message = json.loads(self.webhook_1.main_queue.dequeue())
            self.assertIn(f"<@&{self.group_1.pk}>", message["content"])
            self.assertIn("@here", message["content"])

        def test_can_handle_error_from_discord(self, mock_import_discord_user):
            mock_import_discord_user.return_value.objects.group_to_role.side_effect = (
                HTTPError
            )
            self.tracker.ping_groups.add(self.group_1)

            killmail = self.tracker.process_killmail(load_killmail(10000101))
            self.tracker.generate_killmail_message(
                Killmail.from_json(killmail.asjson())
            )

            self.assertTrue(
                mock_import_discord_user.return_value.objects.group_to_role.called
            )
            self.assertEqual(self.webhook_1.main_queue.size(), 1)
            message = json.loads(self.webhook_1.main_queue.dequeue())
            self.assertNotIn(f"<@&{self.group_1.pk}>", message["content"])



from django.urls import path

from . import views

app_name = "killtracker"

urlpatterns = [
    path(
        "admin_killtracker_toogle_npc/<int:object_id>/",
        views.admin_killtracker_toogle_npc,
        name="admin_killtracker_toogle_npc",
    ),
    path(
        "admin_killtracker_toogle_npc/",
        views.admin_killtracker_toogle_npc,
        name="admin_killtracker_toogle_npc",
    ),
]



# Generated by Django 4.0.3 on 2022-03-02 21:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        ("authentication", "0019_merge_20211026_0919"),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("eveonline", "0015_factions"),
        ("killtracker", "0003_add_state_clauses"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tracker",
            name="exclude_attacker_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="exclude killmails with attackers from one of these alliances",
                related_name="+",
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="exclude_attacker_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="exclude killmails with attackers from one of these corporations",
                related_name="+",
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="exclude_attacker_states",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="exclude killmails with characters belonging to users with these Auth states",
                related_name="+",
                to="authentication.state",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="ping_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Option to ping specific group members - ",
                related_name="+",
                to="auth.group",
                verbose_name="group pings",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attacker_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails with attackers from one of these alliances",
                related_name="+",
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attacker_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails with attackers from one of these corporations",
                related_name="+",
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attacker_states",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails with characters belonging to users with these Auth states",
                related_name="+",
                to="authentication.state",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attackers_ship_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where at least one attacker is flying one of these ship groups",
                related_name="+",
                to="eveuniverse.evegroup",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_attackers_ship_types",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where at least one attacker is flying one of these ship types",
                related_name="+",
                to="eveuniverse.evetype",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_constellations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="+",
                to="eveuniverse.eveconstellation",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_regions",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="+",
                to="eveuniverse.everegion",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_solar_systems",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails that occurred in one of these regions",
                related_name="+",
                to="eveuniverse.evesolarsystem",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_alliances",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails where the victim belongs to one of these alliances",
                related_name="+",
                to="eveonline.eveallianceinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_corporations",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails where the victim belongs to one of these corporations",
                related_name="+",
                to="eveonline.evecorporationinfo",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_ship_groups",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where victim is flying one of these ship groups",
                related_name="+",
                to="eveuniverse.evegroup",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_ship_types",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="Only include killmails where victim is flying one of these ship types",
                related_name="+",
                to="eveuniverse.evetype",
            ),
        ),
        migrations.AlterField(
            model_name="tracker",
            name="require_victim_states",
            field=models.ManyToManyField(
                blank=True,
                default=None,
                help_text="only include killmails where the victim characters belong to users with these Auth states",
                related_name="+",
                to="authentication.state",
            ),
        ),
    ]



# Generated by Django 4.0.7 on 2022-09-03 17:44

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("eveuniverse", "0007_evetype_description"),
        ("killtracker", "0006_evetypeplus"),
    ]

    operations = [
        migrations.AddField(
            model_name="evekillmail",
            name="alliance",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="character",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="damage_taken",
            field=models.BigIntegerField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="faction",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="fitted_value",
            field=models.FloatField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="hash",
            field=models.CharField(blank=True, default="", max_length=64),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="is_awox",
            field=models.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="is_npc",
            field=models.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="is_solo",
            field=models.BooleanField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="location_id",
            field=models.PositiveIntegerField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="position_x",
            field=models.FloatField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="position_y",
            field=models.FloatField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="position_z",
            field=models.FloatField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="ship_type",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="eveuniverse.eveentity",
            ),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="total_value",
            field=models.FloatField(blank=True, db_index=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name="evekillmail",
            name="zkb_points",
            field=models.PositiveIntegerField(
                blank=True, db_index=True, default=None, null=True
            ),
        ),
    ]
